% this code is inspired by VOCevaldet in the PASVAL VOC devkit
function [ap recall precision] = eval_detection(predict_file,gtruth_dir,meta_file,eval_file,blacklist_file,optional_cache_file)
% Evaluate detection
% - predict_file: each line is a single predicted object in the
%   format
%    <image_id> <ILSVRC2013_DET_ID> <confidence> <xmin> <ymin> <xmax> <ymax>
% - gtruth_dir: a path to the directory of ground truth information,
%   e.g., ILSVRC2013_bbox_val/
% - meta_file: information about the synsets
% - eval_file: list of images to evaluate on
% - blacklist_file: list of image/category pairs which aren't
% considered in evaluation
% - optional_cache_file: to save the ground truth data and avoid
% loading from scratch again

if nargin < 3
    meta_file = '../data/meta_det.mat';
end
if nargin < 4
    eval_file = '../data/det_lists/val.txt';
end
if nargin < 5
    blacklist_file = '../data/ILSVRC2013_det_validation_blacklist.txt';
end
if nargin < 6
    optional_cache_file = ''; 
end

defaultIOUthr = 0.5;
pixelTolerance = 10;

load(meta_file);
hash = make_hash(synsets);

if ~isempty(optional_cache_file) && exist(optional_cache_file,'file')
    fprintf('eval_detection :: loading cached ground truth\n');
    load(optional_cache_file);
else
    fprintf('eval_detection :: loading ground truth\n');

    [img_basenames gt_img_ids] = textread(eval_file,'%s %d');

    gt_obj_img_ids = [];
    gt_obj_labels = [];
    gt_obj_bboxes = [];
    t = tic;
    for i=1:length(img_basenames)
        if toc(t) > 10
            fprintf('  eval_detection :: loading on %i of %i\n',...
                    i,length(img_basenames));
            t = tic;
        end
        rec = VOCreadxml(sprintf('%s/%s.xml',gtruth_dir, ...
                                 img_basenames{i}));
        if ~isfield(rec.annotation,'object')
            continue;
        end
        for j=1:length(rec.annotation.object)
            obj = rec.annotation.object(j);
            gt_obj_img_ids(end+1) = gt_img_ids(i);
            gt_obj_labels(end+1) = ...
                get_class2node(hash, obj.name);
            b = obj.bndbox;
            bb = str2double({b.xmin b.ymin b.xmax b.ymax});
            gt_obj_bboxes(:,end+1) = bb;
        end
    end
    gt_w = (gt_obj_bboxes(3,:)-gt_obj_bboxes(1,:)+1);
    gt_h = (gt_obj_bboxes(4,:)-gt_obj_bboxes(2,:)+1);
    thr = (gt_w.*gt_h)./((gt_w+pixelTolerance).*(gt_h+pixelTolerance));
    gt_obj_thr = min(defaultIOUthr,thr);

    num_classes = max(gt_obj_labels);
    num_pos_per_class = zeros(1,num_classes);
    for i=1:num_classes
        num_pos_per_class(i) = sum(gt_obj_labels == i);
    end

    if ~isempty(optional_cache_file)
        save(optional_cache_file,'gt_img_ids','gt_obj_img_ids','gt_obj_labels',...
             'gt_obj_bboxes','gt_obj_thr','num_pos_per_class');
    end
end

blacklist_img_id = [];
blacklist_label = [];
if ~isempty(blacklist_file) && exist(blacklist_file,'file')
    [blacklist_img_id wnid] = textread(blacklist_file,'%d %s');
    blacklist_label = zeros(length(wnid),1);
    for i=1:length(wnid)
        blacklist_label(i) = get_class2node(hash,wnid{i});
    end
    fprintf('eval_detection :: blacklisted %i image/object pairs\n',length(blacklist_label));
end    

fprintf('eval_detection :: loading predictions\n');

[img_ids obj_labels obj_confs xmin ymin xmax ymax] = ...
    textread(predict_file,'%d %d %f %f %f %f %f');
obj_bboxes = [xmin ymin xmax ymax]';

num_pred = length(obj_labels);

[obj_confs ind] = sort(obj_confs,'descend');
img_ids = img_ids(ind);
obj_labels = obj_labels(ind);
obj_bboxes = obj_bboxes(:,ind);

fprintf('eval_detection :: evaluating\n');

tp_all = zeros(1,num_pred);
fp_all = zeros(1,num_pred);

num_classes = length(num_pos_per_class);

t = tic;
% iterate over images
for i=1:length(gt_img_ids)
    if toc(t) > 10
        fprintf('  eval_detection :: evaluating on %i of %i\n',...
                i,length(gt_img_ids));
        t = tic;
    end

    id = gt_img_ids(i);

    bSameImg = gt_obj_img_ids == id;
    gt_labels = gt_obj_labels(bSameImg);
    gt_bboxes = gt_obj_bboxes(:,bSameImg);
    gt_thr = gt_obj_thr(bSameImg);
    num_gt_obj = length(gt_labels);
    gt_detected = zeros(1,num_gt_obj);
   
    bSameImg = blacklist_img_id == id;
    blacklisted_obj = blacklist_label(bSameImg);

    bSameImg = img_ids == id;
    labels = obj_labels(bSameImg);
    bboxes = obj_bboxes(:,bSameImg);
    num_obj = length(labels);
    tp = zeros(1,num_obj);
    fp = zeros(1,num_obj);

    for j=1:num_obj
        if any(labels(j) == blacklisted_obj)
            continue; % just ignore this detection
        end
        bb = bboxes(:,j);
        ovmax = -inf;
        kmax = -1;
        for k=1:num_gt_obj
            if labels(j) ~= gt_labels(k)
                continue;
            end
            if gt_detected(k) > 0
                continue;
            end
            bbgt = gt_bboxes(:,k);
            bi=[max(bb(1),bbgt(1)) ; max(bb(2),bbgt(2)) ; min(bb(3),bbgt(3)) ; min(bb(4),bbgt(4))];
            iw=bi(3)-bi(1)+1;
            ih=bi(4)-bi(2)+1;
            if iw>0 & ih>0                
                % compute overlap as area of intersection / area of union
                ua=(bb(3)-bb(1)+1)*(bb(4)-bb(2)+1)+...
                   (bbgt(3)-bbgt(1)+1)*(bbgt(4)-bbgt(2)+1)-...
                   iw*ih;
                ov=iw*ih/ua;
                % makes sure that this object is detected according
                % to its individual threshold
                if ov >= gt_thr(k) && ov > ovmax
                    ovmax=ov;
                    kmax=k;
                end
            end
        end
        if kmax > 0
            tp(j) = 1;
            gt_detected(kmax) = 1;
        else
            fp(j) = 1;
        end
    end

    % put back into global vector
    tp_all(bSameImg) = tp;
    fp_all(bSameImg) = fp;

    for k=1:num_gt_obj
        label = gt_labels(k);
        % remove blacklisted objects from consideration as positive examples
        if any(label == blacklisted_obj) 
            num_pos_per_class(label) = num_pos_per_class(label)-1;
        end
    end
end

for c=1:num_classes
    % compute precision/recall
    tp = cumsum(tp_all(obj_labels==c));
    fp = cumsum(fp_all(obj_labels==c));
    recall{c}=(tp/num_pos_per_class(c))';
    precision{c}=(tp./(fp+tp))';
    ap(c) =VOCap(recall{c},precision{c});
end

